from torch.utils import data
import h5torch
import numpy as np


class Dataset(data.Dataset):
    def __init__(self, path, sampling = 0):
        self.f = h5torch.File(path)
        if "central" not in self.f:
            raise ValueError("\"central\" data object was not found in input file.")
        if (sampling != "coo") and not isinstance(sampling, int):
            raise TypeError('`sampling` should be either "coo" or `int`')
        if (sampling == "coo") and (self.f["central"].attrs["mode"] not in ["coo", "N-D"]):
            raise ValueError("`coo` sampling only works with central objects stored in `coo` or `N-D` mode")
        if isinstance(sampling, int) and (sampling >= len(self.f["central"].attrs["shape"])): # type: ignore
            raise ValueError("given sampling axis exceeds the number of axes in central data object")
        if isinstance(sampling, int) and (sampling > 0) and (self.f["central"].attrs["mode"] == "csr"):
            raise ValueError("sampling axis can maximally be `0` for central objects in `csr` mode ")
        self.sampling = sampling

    def __len__(self):
        if (self.sampling == "coo") and (self.f["central"].attrs["mode"] == "coo"):
            return self.f["central"]["data"].shape[0] # type: ignore
        elif (self.sampling == "coo") and (self.f["central"].attrs["mode"] == "N-D"):
            return self.f["central"].size # type: ignore
        else:
            return self.f["central"].attrs["shape"][self.sampling] # type: ignore
    def __getitem__(self, index):
        sample = {}

        if self.sampling == "coo":
            if self.f["central"].attrs["mode"] == "N-D":
                indices = np.unravel_index(index, self.f["central"].attrs["shape"]) # type: ignore
                sample["central"] = self.f["central"][indices] # type: ignore
            else:
                indices, data = sample_coo(self.f["central"], index)
                sample["central"] = data
            for axis, ix in enumerate(indices):
                sample |= self.sample_axis(axis, ix, self.f[str(axis)].keys()) # type: ignore
            return sample
        
        else:
            sample["central"] = np.take(self.f["central"], index, self.sampling) # type: ignore
            sample |= self.sample_axis(self.sampling, index, self.f[str(self.sampling)].keys()) # type: ignore
        return sample

    def sample_axis(self, axis, index, keys):
        sample = {}
        for key in keys:
            h5object = self.f["%s/%s" % (axis, key)]
            sample["%s/%s" % (axis, key)] = mode_to_sampler[h5object.attrs["mode"]](h5object, index) # type: ignore
        return sample


    np.unravel_index(15, (2,4,5))

sample_default = lambda h5object, index : h5object[index]
sample_separate = lambda h5object, index : h5object[str(index)][()]
def sample_csr(h5object, index):
    ix0, ix1 = h5object["indptr"][index:index+2]
    x = np.zeros(h5object.attrs["shape"][1], dtype= h5object["data"].dtype)
    x[h5object["indices"][ix0:ix1]] = h5object["data"][ix0:ix1]
    return x
def sample_coo(h5object, index):
    indices = tuple(h5object["indices"][:, index])
    data = h5object["data"][index]
    return indices, data

def sample_ND_as_coo(h5object, index):
    np.unravel_index(index, h5object.attrs["shape"])

mode_to_sampler = {
    "N-D": sample_default,
    "csr": sample_csr,
    "coo": sample_coo,
    "vlen": sample_default,
    "separate": sample_separate,

}


import sys
from setuptools import setup, find_packages

sys.path[0:0] = ["h5torch"]
from version import __version__

from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name="h5torch",
    python_requires=">3.9.0",
    packages=find_packages(),
    version=__version__,
    license="LICENSE.txt",
    description="h5torch",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Gaetan De Waele",
    author_email="gaetandewaele@hotmail.com",
    url="https://github.com/gdewael/h5torch",
    install_requires=[
        "numpy",
        "h5py",
        "torch",
        "scipy"

    ],
)


from scipy import sparse
import h5py
import numpy as np


class File(h5py.File):
    def __init__(self, path, mode = "r"):
        super().__init__(path, mode)

    def register(self, data, axis, name = None, mode = "N-D", dtype = None):
        """
        Note: if mode = "coo": then data should be a tuple consisting of indices, data and a shape.
        If it's a np.ndarray, it will be first converted to a 2D scipy.sparse.coo_matrix
        """
        if isinstance(name, str) and ("/" in name): raise ValueError("\"/\" not allowed in name as it creates a different group structure")
        if mode not in ["N-D", "csr", "coo", "vlen", "separate"]:
            raise ValueError('mode not in ["N-D", "csr", "coo", "vlen", "separate"]')
        if (axis not in ["central", "unstructured"]) and not isinstance(axis, int):
            raise TypeError('axis should be an `int` or "central" or "unstructured')
        if axis != "central" and "central" not in self:
            raise AssertionError("\"central\" data object should exist before other items are added")
        if isinstance(axis, int) and (axis >= len(self["central"].attrs["shape"])): # type: ignore
            raise ValueError("given alignment axis exceeds the number of axes in central data object")
        if mode == "csr": data = sparse.csr_matrix(data)
        len_ = (len(data) if not isinstance(data, sparse.csr_matrix) else data.shape[0])
        if mode != "coo" and isinstance(axis, int) and (len_ != self["central"].attrs["shape"][axis]):  # type: ignore
            raise ValueError(
                """the number of rows in the given data does not equal the number of elements in the
                central data object along its alignment axis.""")
        if axis != "central" and name is None: 
            raise TypeError("`name` should be specified for data objects that are not the central object.")
        if (mode == "coo") and (axis not in ["central", "unstructured"]):
            raise ValueError("COO sparse matrix not supported for `axis` data objects.")

        if mode == "N-D": register_fun = self._ND_register
        elif mode == "csr": register_fun = self._csr_register
        elif mode == "coo": register_fun = self._coo_register
        elif mode == "vlen": register_fun = self._vlen_register
        elif mode == "separate": register_fun = self._separate_register

        if axis == "central":
            name = "central"
        else:
            name = "%s/%s"% (axis, name)
    

        register_fun(data, name, dtype) # type: ignore

        # data = data.astype(np.dtype(dtype)))

    def __repr__(self):
        f = h5py.File.__repr__(self)
        return f[0] + "h5torch" + f[5:]

    def _ND_register(self, data, name, dtype):
        self.create_dataset(name, data = data)
        self[name].attrs["shape"] = data.shape
        self[name].attrs["mode"] = "N-D"

    def _csr_register(self, data, name, dtype):
        self.create_dataset("%s/data" % name, data = data.data)
        self.create_dataset("%s/indices" % name, data = data.indices)
        self.create_dataset("%s/indptr" % name, data = data.indptr)
        self[name].attrs["shape"] = data.shape
        self[name].attrs["mode"] = "csr"

    def _coo_register(self, data, name, dtype):
        if isinstance(data, np.ndarray):
            data = sparse.coo_matrix(data)
            data = (np.stack([data.row, data.col]), data.data, data.shape)
            
        self.create_dataset("%s/indices" % name, data = data[0])
        self.create_dataset("%s/data" % name, data = data[1])
        self[name].attrs["shape"] = data[2]
        self[name].attrs["mode"] = "coo"


    def _vlen_register(self, data, name, dtype):
        if not all([len(elem.shape)==1 for elem in data]):
            raise ValueError("All elements in data should be 1D for `mode=vlen`")
        
        self.create_dataset(name, data = data, dtype = h5py.vlen_dtype(data[0].dtype))
        self[name].attrs["shape"] = (len(data), )
        self[name].attrs["mode"] = "vlen"


    def _separate_register(self, data, name, dtype):
        for ix, elem in enumerate(data):
            self.create_dataset("%s/%s" % (name, ix), data = elem)

        self[name].attrs["shape"] = (len(data), )
        self[name].attrs["mode"] = "separate"

